#include <iostream>
#include <iomanip>
#include <fstream>
#include <string>
#include <cassert>
#include <cmath>

#include "vector.h"
#include "pixel.h"
#include "camera.h"
#include "light.h"
#include "material.h"
#include "object.h"
#include "list.h"
#include "plane.h"
#include "model.h"
#include "ray.h"

bool ray_t::feeler(model_t&		model,
                   rgb_t<double>&	color,
                   int			bounce)
{
	// (static) object properties
	object_t			*obj=NULL;
	vec_t				hit, N;

  // get closest object, if any
  if(!(obj = model.find_closest(pos,dir,dis,hit,N)) || 0>dis || dis>MAX_DIST)
    return false;

  return true;
}

void ray_t::trace(model_t&		model,
                  rgb_t<double>&	color,
                  int			bounce)
{
	// (static) object properties
	object_t			*obj=NULL;
	rgb_t<double>			ambient, diffuse, specular;
	material_t			*mat;
	vec_t				hit, N;

	// light and illumination model variables
	light_t				*lgt=NULL;
	vec_t				L,V,R,H;
	rgb_t<double>			I_d, I_s;
	double				r,ndotl=0.0,n=32.0;
	double				alpha=0.0,ior;
	list_t<light_t* >::iterator	litr;

  if(bounce > RAY_BOUNCES) return;

  // get closest object, if any
  if(!(obj = model.find_closest(pos,dir,dis,hit,N)) || dis > MAX_DIST)
    return;

  // if hit distance valid, compute color at surface
  if(dis > 0) {

    // get object material properties
    if((mat = model.getmaterial(obj->getmaterial())) != NULL) {
      ambient = mat->getamb();
      diffuse = mat->getdiff();
      specular = mat->getspec();
      alpha = mat->getalpha();
      ior = mat->getior();
    }

    // ambient color
    color += 1.0/dis * ambient;			// ambient scaled by ray dist

    // clamp resultant color
    color.clamp(0.0,1.0);

    // view direction (direction from hit point to camera)
    V = -dir;				// dir to camera

    // diffuse component from each light...
    if(!diffuse.iszero()) {
      for(litr = model.lgts.begin(); litr != model.lgts.end(); litr++) {
        // pointer to light
        lgt = (light_t *)*litr;
  
        // light direction and distance
        L = lgt->getlocation() - hit;
        r = L.len();				// distance to light
        L = L.norm();				// dir to light
  
        // angle with light
        ndotl = N.dot(L);
  
        // check visibility wrt light
        if(0.0 < ndotl && ndotl < 1.0) {
          // light color scaled by N . L
          I_d = 1.0/r * ndotl * lgt->getcolor();
  
          // add in diffuse contribution from light scaled by material property
          color += (I_d * diffuse);
        }
      }
  
      // clamp resultant color
      color.clamp(0.0,1.0);
    }

    // reflection
    if(!specular.iszero()) {
	rgb_t<double>	refcolor;
	vec_t		r = dir.reflect(N);
	ray_t		*reflection = new ray_t(hit,r,dis);

      // trace the reflection
      reflection->trace(model,refcolor,bounce+1);
      delete reflection;

      // composite surface color: blending baesd on surface properties
      // (note that diffuse + specular should add up to 1)
      color  = (diffuse * color) + (specular * refcolor);

      // clamp resultant color
      color.clamp(0.0,1.0);
    }

    // transmission, see: http://www.cs.utah.edu/~shirley/books/fcg2/rt.pdf
    if(alpha > 0) {
	double		ndotd = dir.dot(N);
	rgb_t<double>	refcolor,trncolor;
	vec_t		r = dir.reflect(N);
	vec_t		t = ndotd < 0 ? dir.refract(N,ior) :	// ray in
			              dir.refract(-N,1.0/ior);	// ray out
	double		c = ndotd < 0 ? N.dot(-dir) : N.dot(t);
	ray_t		*reflection = new ray_t(hit,r,dis);
	ray_t		*transmission = new ray_t(hit,t,dis);
	double		R,R_0 = ((ior-1.0)*(ior-1.0))/((ior+1.0)*(ior+1.0));

      // trace the reflection
      reflection->trace(model,refcolor,bounce+1);
      delete reflection;

      // trace the transmission
      transmission->trace(model,trncolor,bounce+1);
      delete transmission;

      // Schlick's approximation to the Fresnel equations
      R = R_0 + (1.0 - R_0)*pow(1.0 - c,5.0);

      // composite surface color: blending baesd on surface properties
      // (note that diffuse + specular should add up to 1)
    //color  = ((1.0 - alpha) * color) + (alpha * trncolor);
      color  = ((1.0 - alpha) * color) +
                      (alpha * ((1.0 - R) * trncolor + (R * refcolor)));

      // clamp resultant color
      color.clamp(0.0,1.0);
    }

    // specular highlights
    if(!specular.iszero()) {

      // add in specular highlight from each light...
      for(litr = model.lgts.begin(); litr != model.lgts.end(); litr++) {
        // pointer to light
        lgt = (light_t *)*litr;

        // light direction and distance
        L = lgt->getlocation() - hit;
        r = L.len();				// distance to light
        L = L.norm();				// dir to light

        // angle with light
        ndotl = N.dot(L);

        // check visibility wrt light
        if(0.0 < ndotl && ndotl < 1.0) {
          // specular reflection R = (2.0 * (N . L) * N) - L
          R = L.reflect(N);
          // bisector (not used)
          H = (0.5 * (L + V)).norm();

          // light color scaled by (R . V)^n
          I_s = 1.0/r * pow(R.dot(V),n) * lgt->getcolor();
  
          // add in specular contribution from light scaled by material property
          color += (I_s * specular);
        }

        // clamp resultant color
        color.clamp(0.0,1.0);
      }
    }

/*
    // simple shadown casting
    for(litr = model.lgts.begin(); litr != model.lgts.end(); litr++) {

      // pointer to light
      lgt = (light_t *)*litr;

      // light direction and distance
      L = lgt->getlocation() - hit;
      r = L.len();				// distance to light
      L = L.norm();				// dir to light

	ray_t	*shadow = new ray_t(hit,L,0);

      if(shadow->feeler(model,color,0))
        color *= 0.9;
      delete shadow;
    }

    // clamp resultant color
    color.clamp(0.0,1.0);
*/

  }
}
